/*
 * 85_mux_ch_dec.c
 *
 *  ATtiny85 como decodificador, con multiplexacin de leds
 *  en conexin charlieplexing de cuatro pines
 *
 * Created: 15/04/2014 10:09:05
 *  Author: F. Germn Trinidad B.
 */

#include <avr/interrupt.h>
#include <avr/eeprom.h>

#define FCPU 8000000UL

// pines empleados. PB4 se usa para PCINT0
#define pin1 PB0
#define pin2 PB1
#define pin3 PB2
#define pin4 PB3

// velocidad loco
#define VEL_BAJA 10
#define VEL_ALTA 18

// Lmites brillo
#define BRILLO_MAX 16
#define BRILLO_MIN  2

// tipos:
#define INC_DP 0        // incandescente dptos
#define FLR_DP 1        // fluorescente dptos
#define INC_SL 2        // incandescente saln
#define FLR_SL 3        // fluorescente saln

// CVs en eeprom:
#define CV_DIR 0        // [P]+[*]              CV1
#define CV_LOCO 1       // [P]+[F1] vel min     CV2
#define CV_BRILLO 2     // [P]+[F2] acel        CV3
#define CV_TIPO 3       // [P]+[F3] decel       CV4
#define CV_RND 4        // [P]+[F4] vel max     CV5
// [P]+[STOP] CV29

// Funciones DCC:
// vienen en el comando:  '1 0 0 F0   F4 F3 F2 F1'
#define F0 16
#define F1  1
#define F2  2
#define F3  4
#define F4  8

typedef unsigned char byte;

// de recepcin DCC
byte num_bits;              // nmero de bits ledos
byte est_DCC;               // 0->pre 1->drc 2->cmd 3->xor
volatile byte drc_dcc;      // recibidos
volatile byte cmd_dcc;      // funciones recibidas
volatile byte vel_loco;     // velocidad
volatile byte modo_prg;     // indicador de modo programacin
byte bit_dcc;               // bit recibido

// de la eeprom (en defines sus direcciones en eeprom):
byte cv_dir;                // direccin propia (valor 2)
byte cv_loco;               // direccin de la locomotora (valor 3)
byte cv_brillo;             // nivel de iluminacin (valor 16)
byte cv_tipo;               // tipo de iluminacin
byte cv_rnd;                // valor para diferenciar los coches

// generales
volatile unsigned int leds;
volatile unsigned int ct_us;    // contador para 100 us con timer0
volatile byte nro_rnd;          // nmero de flancos de bits DCC recibidos (pseudoaleatorio)

// macros:
#define pinH(x) PORTB |= (1<<x)         // pin x a nivel alto
#define pinL(x) PORTB &= ~(1<<x)        // ... bajo
#define pin(x,y,z) ((z!=0) ? (x |= (1<<y)) : (x &= ~(1<<y)))

// refresco de los LEDs, duracin 100 ms
void mux ()
{
    // poniendo cada LED
    for (byte repet=5; repet>0; repet--)    // repite 5 veces cada refresco de 20 ms = 100 ms
    {   // si el LED va a estar iluminado
        if(leds & (1<<0)) {DDRB = 0b00000011; PORTB = 0b00000010;}      // D1   pin1-pin2
        ct_us = cv_brillo;          // carga el contador de la cuenta
        while(ct_us);               // espera a terminar la cuenta en cualquier caso
        DDRB=0;                     //  y vuelve a poner los pines como entradas
        PORTB = 0;                  //  sin pullup

        if(leds & (1<<1))  {DDRB = 0b00000011; PORTB = 0b00000001;}     // D2   pin2-pin1
        ct_us = cv_brillo; while(ct_us); DDRB=0; PORTB = 0;             // para facilitar la lectura

        if(leds & (1<<2))  {DDRB = 0b00000110; PORTB = 0b00000100;}      // D3   pin2-pin3
        ct_us = cv_brillo; while(ct_us); DDRB=0; PORTB = 0;

        if(leds & (1<<3))  {DDRB = 0b00000110; PORTB = 0b00000010;}      // D4   pin3-pin2
        ct_us = cv_brillo; while(ct_us); DDRB=0; PORTB = 0;

        if(leds & (1<<4))  {DDRB = 0b00001100; PORTB = 0b00001000;}      // D5   pin3-pin4
        ct_us = cv_brillo; while(ct_us); DDRB=0; PORTB = 0;

        if(leds & (1<<5))  {DDRB = 0b00001100; PORTB = 0b00000100;}      // D6   pin4-pin3
        ct_us = cv_brillo; while(ct_us); DDRB=0; PORTB = 0;

        if(leds & (1<<6))  {DDRB = 0b00000101; PORTB = 0b00000100;}      // D7   pin1-pin3
        ct_us = cv_brillo; while(ct_us); DDRB=0; PORTB = 0;

        if(leds & (1<<7))  {DDRB = 0b00000101; PORTB = 0b00000001;}      // D8   pin3-pin1
        ct_us = cv_brillo; while(ct_us); DDRB=0; PORTB = 0;

        if(leds & (1<<8))  {DDRB = 0b00001010; PORTB = 0b00001000;}      // D9   pin2-pin4
        ct_us = cv_brillo; while(ct_us); DDRB=0; PORTB = 0;

        if(leds & (1<<9))  {DDRB = 0b00001010; PORTB = 0b00000010;}      // D10   pin4-pin2
        ct_us = cv_brillo; while(ct_us); DDRB=0; PORTB = 0;

        if(leds & (1<<10)) {DDRB = 0b00001001; PORTB = 0b00001000;}      // D11   pin1-pin4
        ct_us = cv_brillo; while(ct_us); DDRB=0; PORTB = 0;

        if(leds & (1<<11)) {DDRB = 0b00001001; PORTB = 0b00000001;}      // D12   pin4-pin1
        ct_us = cv_brillo; while(ct_us); DDRB=0; PORTB = 0;

        ct_us = 12*BRILLO_MAX - cv_brillo*12;                   // tiempo restante hasta unos 20 ms (50Hz)
        while(ct_us);
    }
}

// Interrupcin externa para la seal DCC
// Cuenta el tiempo del nivel alto con el timer1
ISR(PCINT0_vect)
{
    byte tmp_pulso;
    static byte drc, cmd, xor;                      // los que se forman aqu
    if (PINB & (1<<PB4)) TCCR1 |= 1<<CS12;          // cuando se pone a nivel alto, arranca cuenta prescaler 1/8
    else                                            // cuando pasa a nivel bajo
    {
        TCCR1 &= ~(1<<CS12);                        // detiene la cuenta
        tmp_pulso = TCNT1;                          // lee los microsegundos
        TCNT1 = 0;                                  // pone a cero el timer
        if (tmp_pulso>90) bit_dcc = 0;              // sin buscar exactitud: duracin mayor de 90 us => cero
        else bit_dcc = 1;                           //    y menor uno
        if (!est_DCC)                               // prembulo
        {
            if (bit_dcc)                            // otro '1'
            {if (num_bits) num_bits--;}
            else
            {
                if (!num_bits)                      // el '0' de inicio de datos
                {est_DCC = 1; num_bits = 9;}
                else                                // no se cumple la trama, vuelta a inicio
                num_bits = 10;
            }
        }
        else if (est_DCC == 1)
        {
            if (--num_bits)                         // bit de direccin
            {drc *= 2; if (bit_dcc) drc++;}         // se recibi un '1'
            else
            {   // si es el '0' de inicio de datos y direccin correcta (decoder o loco) o prg
                if (!bit_dcc )
                {
                    if (drc!=0x7c)                      // cualquier direccin
                    {est_DCC = 2; num_bits = 9;}        //  pasa al siguiente
                    else                                // entra en modo programacin
                    {modo_prg = 1; est_DCC = 1; num_bits = 9;}   // indica modo de programacin y vuelve para recibir direccin de CV
                }
                else                                    // en otro caso, vuelta a inicio
                {est_DCC = 0; num_bits = 10;}
            }
        }
        else if (est_DCC == 2)
        {
            if (--num_bits)                             // bit de comando
            {cmd *= 2;if (bit_dcc) cmd++;}              // se recibi un '1'
            else
            {
                if (!bit_dcc)                           // el '0' de inicio de datos
                {est_DCC = 3; num_bits = 9;}
                else                                    // no se cumple la trama, vuelta a inicio
                {est_DCC = 0; num_bits = 10;}
            }
        }
        else                        // estado == 3
        {
            if (--num_bits)                             // bit de xor
            {xor *= 2; if (bit_dcc) xor++;}             // se recibi un '1'
            else
            {
                if (bit_dcc)    // el '1' de fin de la trama
                {
                    if (!modo_prg)
                    {
                        if (xor == (drc^cmd))
                        {   // slo vale el comando de funciones de la direccin del decoder
                            if ((drc==cv_dir) && (cmd & 0xe0) == 0x80)
                            cmd_dcc = cmd;                                  // funciones: 100x xxxx del decod
                            else if ((drc==cv_loco) && (cmd & 0xc0) == 0x40) // de velocidad: 01XD DDDD de la loco
                            {
                                vel_loco = (cmd & 0x0f)*2;                  // 28 pasos de velocidad
                                if ((cmd & 0x10) == 0x10) vel_loco++;       // se aade el bit de menor peso
                                if (vel_loco>3) vel_loco -= 3;              // para dejar los pasos de 1 a 28
                                else vel_loco = 0;                          //  y si no, 0 para las luces
                            }
                        }
                    }
                    else    // est en modo programacin
                    {
                        if (xor == (drc^cmd^0x7c))                          // los tres bytes recibidos
                        {
                            cli();                                          // impide interr
                            // graba el comando en la eeprom:
                            if (drc==CV_DIR) eeprom_update_byte((uint8_t*) CV_DIR, cmd);
                            else if (drc==CV_BRILLO) eeprom_update_byte((uint8_t*) CV_BRILLO, cmd);
                            else if (drc==CV_LOCO) eeprom_update_byte((uint8_t*) CV_LOCO, cmd);
                            else if (drc==CV_TIPO) eeprom_update_byte((uint8_t*) CV_TIPO, cmd);
                            else if (drc==CV_RND) eeprom_update_byte((uint8_t*) CV_RND, cmd);
                            else while (1);             // si no son estas no hace nada
                            GIMSK &= ~(1<<PCIE);        // impide interrumpir a sta
                            sei();
                            // en cambio si ha grabado, hace un Ack
                            leds = 0x0fff;                      // enciende todos los LEDs
                            for (byte i=5; i>0; i--) mux();     // Ack "visual"
                            while (1);                          // hasta que desconecte el programador
                        }
                    }
                    est_DCC = 0;                                // como era fin de la trama, vuelta a empezar
                    num_bits = 10;
                }
            }
        }
    }
    nro_rnd++;              // buscando un nmero 'aleatorio'
}

// Interrupcin del timer0 para el refresco
ISR(TIMER0_COMPA_vect, ISR_NOBLOCK)
{
    if (ct_us) ct_us--;                                 // slo cuenta cuando se le carga
}

// devuelve un nmero segn la funcin a realizar
byte ver_estado()
{
byte retornar=0;
    switch (cmd_dcc & 0b00011111)       // funciones previstas
    {
        case 0:  break;
        case F0: retornar = 1; break;       // encendido todos
        case F0+F1: retornar = 2; break;    // control de nivel de iluminacin
        case F2: retornar = 3; break;       // encendido aleatorio
        case F3: retornar = 4; break;       // apagado aleatorio
        case F4: retornar = 5; break;       // encendido/apagado aleatorio
        case F0+F4:                         // pasa el control a la velocidad de la locomotora
            if (cv_tipo == INC_DP || cv_tipo == FLR_DP)     // slo para el caso de departamentos
            {
                if (vel_loco > VEL_ALTA) retornar = 4;      // apaga departamentos en velocidad elevada
                else if (vel_loco > VEL_BAJA) retornar = 5; // los enciende al reducir la velocidad
                else retornar = 3;                          // enciende/apaga en velocidades intermedias
            }
    }
    return retornar;
}

// configuracin inicial para ATtiny85
void setup()
{

    DDRB=0;                                             // todo entradas
    GIMSK |= (1<<PCIE);                                 // habilita interrupcin por cambio de nivel
    PCMSK |= (1<<PCINT4);                               // interrupcin por cambio de PB4

    TCCR0A |= (1 << WGM01);                             // timer0 en modo CTC para mux
    OCR0A = 99;                                         // compare... (99 -> 5 KHz => T->200 us => T/2 -> 100 us)
    TCCR0B |= ((1 << CS01));                            // prescaler 1/8
    TIMSK |= (1 << OCIE0A);                             // habilita interrupcin

    est_DCC = 0;                                        // estado inicial de la seal DCC
    num_bits = 10;                                      // nmero mnimo en el prembulo
    modo_prg = 0;                                       // indica que no est en modo programacin
    // lee las variables almacenadas en eeprom
    cv_dir = eeprom_read_byte((uint8_t*) CV_DIR);
    cv_loco = eeprom_read_byte((uint8_t*) CV_LOCO);
    cv_brillo = eeprom_read_byte((uint8_t*) CV_BRILLO);
    // si no est dentro del rango, toma el bsico
    if ((cv_brillo > BRILLO_MAX) || (cv_brillo < BRILLO_MIN)) cv_brillo = BRILLO_MAX;
    cv_tipo = eeprom_read_byte((uint8_t*) CV_TIPO);
    if (cv_tipo > 3) cv_tipo = INC_DP;                  // si error, incandescente por departamentos
    cv_rnd = eeprom_read_byte((uint8_t*) CV_RND);
    if (cv_rnd == 255) cv_rnd = 1;                      // para evitar el 255=-1
    sei();                                              // habilita las interrupciones
}

// encendido de un led fluorescente
void enc_fluor (byte nro_led)
{
    byte parp=0;
    for (byte rep=(nro_rnd % 4)+3; rep>0; rep--)
    {
        if (!parp) leds |= (1<<nro_led);                // en funcin del valor de parp lo enciende
        else leds &= ~(1<<nro_led);                     //  o lo apaga
        parp = ~parp;
        mux();
        if (rep == 2) mux();                            // para hacer un efecto de persistencia al final
    }
    leds |= (1<<nro_led);                               // acaba encendido
}

// F0: enciende todos los LEDs
void enc_todos()
{
    switch (cv_tipo)
    {
        case INC_DP:                    // los incandescentes se encienden inmediatamente
        case INC_SL:
        leds = 0x0fff;
        break;

        case FLR_DP:                    // los fluorescentes parpadean
        case FLR_SL:
        if (leds == 0x0fff) break;      // si ya estn encendidos
        for (byte rep=(nro_rnd % 4)+3; rep>0; rep--)    // un cierto nmero de veces
        {
            leds = 256*nro_rnd + nro_rnd;   // valor imprevisible
            mux();
            if (rep == 1) mux();        // efecto de persistencia al final
        }
        leds = 0x0fff;                  // todos encendidos ya
        break;
    }
}

// F2: encendido aleatorio
void enc_aleat()
{
static byte rep;
static byte nro_led;
    if (leds == 0x0fff) return;             // si estn encendidos
    if (rep) rep--;
    else
    {
        rep = (nro_rnd & 0x1f) + 1;         // nmero de repeticiones (pausa entre encendidos)
        nro_led = (nro_rnd) % 12;           // posicin del LED a encender en su variable
        switch (cv_tipo)                    // segn el tipo de iluminacin
        {
            case INC_DP:                    // por departamentos
            case FLR_DP:
            while(leds & (1<<nro_led))      // mientras ya est encendido
            {
                nro_led++;                  //  busca el primero apagado
                if (nro_led > 11) nro_led -= 12;
            }
            if(cv_tipo == INC_DP) leds |= (1<<nro_led); // lo enciende segn el tipo
            else enc_fluor (nro_led);
            break;
            case INC_SL:                // estos casos no existen
            case FLR_SL:
            break;
        }
    }
}

// F3: apagado aleatorio
void apaga_aleat()
{
static byte rep;
static byte nro_led;
    if (!leds) return;                       // ya estn apagados
    if(cv_tipo == INC_SL || cv_tipo == FLR_SL)
        leds = 0;                           // si es tipo saln, apaga todos
    else
    {
        if (rep) rep--;
        else
        {
            rep = (nro_rnd & 0x1f) + 1;
            nro_led = (nro_rnd) % 12;
            while(!(leds & (1<<nro_led)))
            {
                nro_led++;                  // si est apagado, busca el primero encendido
                if (nro_led > 11) nro_led -= 12;
            }
            leds &= ~(1<<nro_led);          // lo apaga
        }
    }
}

// F4: encendido/apagado
void enc_apg_aleat()
{
static byte nro_led;
static byte rep;
    if (rep) rep--;
    else
    {
        rep = (nro_rnd & 0x1f) + 5;         // ms de un seg
        switch (cv_tipo)                    // slo existen los tipos de departamentos
        {
            case INC_DP:
            leds ^= (1<<(nro_rnd % 12));    // incandescente, cambia de estado
            break;
            case FLR_DP:                    // fluorescente
            nro_led = (nro_rnd) % 12;
            if (leds & (1<<nro_led)) leds &= ~(1<<nro_led );    // estaba encendido, lo apaga
            else enc_fluor(nro_led);                            // y si no, lo enciende fluorescente
        }
    }
}


int main(void)
{
char delta=1;               // para subir/bajar el brillo
// ------------------------------------------------------------------
//  Prueba con estados (dependiendo de funciones y velocidad loco):
//      0 -> todo apagado           1 -> todo encendido
//      2 -> ajuste iluminacin     3 -> encender dptos
//      4 -> apagar dptos           5 -> encendido/apagado aleatorio
//
byte estado=0;

    setup();            // configuracin inicial
    while(1)            // bucle
    {
        nro_rnd += cv_rnd;                              // que sea lo ms aleatorio posible
        estado = ver_estado();                          // consulta los comandos
        switch (estado)
        {
            case 1:         // todos los leds encendidos
                enc_todos ();
                break;
            case 2:         // F0+F1: variacin de brillo
                if (cv_brillo == BRILLO_MAX)
                {delta = -1; mux(); mux();}             // en los lmites se para ms
                else if (cv_brillo == BRILLO_MIN)
                {delta = 1; mux(); mux();}
                cv_brillo += delta;                     // suma o resta 1
                mux();
                break;
            case 3:         // se van encendiendo
                enc_aleat();
                break;
            case 4:         // se van apagando
                apaga_aleat();
                break;
            case 5:         // encienden/apagan aleatorio
                enc_apg_aleat();
                break;
        }
        mux();              // refresco de los LEDs
    }
}
